import math


def threshold_sturges(values, *args):
    values = [v for v in values if v is not None]
    if len(values) == 0:
        return 1
    return max(1, math.ceil(math.log(len(values)) / math.log(2)) + 1)



NAMESPACES = {
    "svg": {"xmlns": "http://www.w3.org/2000/svg"},
    # 'xhtml': xhtml,
    # 'xlink': "http://www.w3.org/1999/xlink",
    # 'xml': "http://www.w3.org/XML/1998/namespace",
    # 'xmlns': "http://www.w3.org/2000/xmlns/"
}


def namespace(name):
    prefix = name + ""
    i = prefix.find(":")
    if i >= 0 and (prefix := name[:i]) != "xmlns":
        name = name[i + 1 :]
    return (
        {"space": NAMESPACES[prefix], "local": name} if prefix in NAMESPACES else name
    )



from __future__ import annotations

import math
from collections.abc import Callable
from datetime import datetime
from typing import Any, TypeVar, overload

from ..interpolate import interpolate, interpolate_round
from .continuous import identity
from .init import init_interpolator
from .linear import LinearBase
from .log import LogBase, logp, powp, reflect, transform_log, transform_logn
from .pow import transform_pow, transform_sqrt
from .symlog import transform_symlog

T = TypeVar("T")


class Sequential:
    def __init__(self, t: Callable):
        self._x0 = 0
        self._x1 = 1
        self._transform = t
        self._t0 = t(self._x0)
        self._t1 = t(self._x1)
        self._k10 = 0 if self._t0 == self._t1 else 1 / (self._t1 - self._t0)
        self._interpolator = identity
        self._clamp = False
        self._unknown = None

    def __call__(self, x: int | float) -> T:
        """
        Given a value from the domain, returns the corresponding value from the range.

        Parameters
        ----------
        x : int | float
            Input value

        Returns
        -------
        T
            Corresponding value from the range
        """
        if x is None or (isinstance(x, float) and math.isnan(x)):
            return self._unknown
        if self._k10 == 0:
            x = 0.5
        else:
            x = (self._transform(x) - self._t0) * self._k10
            if self._clamp:
                x = max(0, min(1, x))
        return self._interpolator(x)

    def set_domain(self, domain: list[int | float]) -> Sequential:
        """
        Sets the scale's domain to the specified array of numbers

        Parameters
        ----------
        domain : list[int | float]
            Domain

        Returns
        -------
        Sequential
            Itself
        """
        self._x0, self._x1 = map(float, list(domain)[:2])
        self._t0 = self._transform(self._x0)
        self._t1 = self._transform(self._x1)
        self._k10 = 0 if self._t0 == self._t1 else 1 / (self._t1 - self._t0)
        return self

    @property
    def domain(self) -> list[int | float]:
        return [self._x0, self._x1]

    def set_clamp(self, clamp: bool) -> Sequential:
        """
        Enables or disables clamping accordingly.

        Parameters
        ----------
        clamp : bool
            Clamp value

        Returns
        -------
        Sequential
            Itself
        """
        self._clamp = bool(clamp)
        return self

    @property
    def clamp(self) -> bool:
        return self._clamp

    def set_interpolator(self, interpolator: Callable) -> Sequential:
        """
        Sets the scale’s interpolator to the specified function.

        Parameters
        ----------
        interpolator : Callable
            Interpolator function

        Returns
        -------
        Sequential
            Itself
        """
        self._interpolator = interpolator
        return self

    @property
    def interpolator(self) -> Callable:
        return self._interpolator

    def set_range(self, range_vals: list[T]) -> Sequential:
        """
        The given two-element array is converted
        to an interpolator function using interpolate

        Parameters
        ----------
        range_vals : list[T]
            Two values

        Returns
        -------
        Sequential
            Itself
        """
        self._r0, self._r1 = list(range_vals)[:2]
        self._interpolator = interpolate(self._r0, self._r1)
        return self

    @property
    def range(self) -> list[T]:
        return [self._interpolator(0), self._interpolator(1)]

    def set_range_round(self, range_vals: list[T]) -> Sequential:
        """
        Sets the scale's range to the specified array of values
        and sets scale's interpolator to :code:`interpolate_round`.

        Parameters
        ----------
        range_vals : list[T]
            Range values

        Returns
        -------
        Sequential
            Itself
        """
        self._r0, self._r1 = range_vals
        self._interpolator = interpolate_round(self._r0, self._r1)
        return self

    @property
    def range_round(self) -> list[T]:
        return [self._interpolator(0), self._interpolator(1)]

    def set_unknown(self, unknown: Any) -> Sequential:
        """
        Sets the output value of the scale for undefined
        or NaN input values.

        Parameters
        ----------
        unknown : Any
            Unknown value

        Returns
        -------
        Sequential
            Itself
        """
        self._unknown = unknown
        return self

    @property
    def unknown(self) -> Any:
        return self._unknown


def copy(source, target):
    return (
        target.set_domain(source.domain)
        .set_interpolator(source.interpolator)
        .set_clamp(source.clamp)
        .set_unknown(source.unknown)
    )


class SequentialLinear(Sequential, LinearBase):
    def __init__(self):
        Sequential.__init__(self, identity)

    def copy(self):
        return copy(self, SequentialLinear())


class SequentialLog(Sequential, LogBase):
    def __init__(self):
        Sequential.__init__(self)
        LogBase.__init__(self)
        self.transform = transform_log
        self.set_domain([1, 10])

    def _rescale(self):
        self._logs = logp(self._base)
        self._pows = powp(self._base)
        d = self.domain[0]
        if isinstance(d, datetime):
            d = d.timestamp()
        if d < 0:
            self._logs = reflect(self._logs)
            self._pows = reflect(self._pows)
            self.transform = transform_logn
        else:
            self.transform = transform_log
        return self

    def set_domain(self, domain: list[int | float]) -> SequentialLog:
        """
        Sets the scale's domain to the specified array of numbers

        Parameters
        ----------
        domain : list[int | float]
            Domain

        Returns
        -------
        SequentialLog
            Itself
        """
        self._x0, self._x1 = map(float, list(domain)[:2])
        self._rescale()
        super().set_domain(domain)
        return self

    def copy(self):
        return copy(self, SequentialLog()).base(self.base)


class SequentialSymlog(Sequential):
    def __init__(self, c: int | float = 1):
        self._c = c
        super().__init__(transform_symlog(self._c))

    def set_constant(self, c: int | float) -> SequentialLog:
        """
        Sets the symlog constant to the specified number and returns this scale.

        Parameters
        ----------
        c : int | float
            Constant value

        Returns
        -------
        SequentialLog
            Itself
        """
        self._c = float(c)
        self.transform = transform_symlog(self._c)
        self.rescale()
        return self

    def copy(self):
        return copy(self, SequentialSymlog()).set_constant(self.constant)


class SequentialPow(Sequential, LinearBase):
    def __init__(self, t: Callable = identity):
        super().__init__(t)
        self._exponent = 1

    def _rescale(self):
        if self._exponent == 1:
            self.transform = identity
            self.rescale()
            return self
        elif self._exponent == 0.5:
            self.transform = transform_sqrt
            self.rescale()
            return self
        else:
            self.transform = transform_pow(self._exponent)
            self.rescale()
            return self

    def set_exponent(self, exponent: int | float) -> SequentialPow:
        """
        Sets the scale's exponent value.

        Parameters
        ----------
        exponent : int | float
            Exponent value

        Returns
        -------
        SequentialPow
            Itself
        """
        self._exponent = float(exponent)
        return self._rescale()

    @property
    def exponent(self):
        return self._exponent

    def copy(self):
        return copy(self, SequentialPow()).set_exponent(self.exponent)


@overload
def scale_sequential() -> SequentialLinear: ...


@overload
def scale_sequential(interpolator: Callable) -> SequentialLinear: ...


@overload
def scale_sequential(
    domain: list[int | float], interpolator: Callable
) -> SequentialLinear: ...


def scale_sequential(*args):
    """
    Builds a new sequential scale with the specified
    domain and interpolator function or array.

    Parameters
    ----------
    domain : list[int | float]
        Domain
    interpolator : Callable
        Interpolator

    Returns
    -------
    SequentialLinear
        Sequential object

    Examples
    --------

    >>> d3.scale_sequential([0, 100], d3.interpolate_blues)
    """
    scale = SequentialLinear()
    if len(args) == 1:
        return init_interpolator(scale, interpolator=args[0])
    elif len(args) == 2:
        domain, interpolator = args
        return init_interpolator(scale, domain=domain, interpolator=interpolator)
    return init_interpolator(scale)


@overload
def scale_sequential_log() -> SequentialLog: ...


@overload
def scale_sequential_log(interpolator: Callable) -> SequentialLog: ...


@overload
def scale_sequential_log(
    domain: list[int | float], interpolator: Callable
) -> SequentialLog: ...


def scale_sequential_log(*args):
    """
    Builds a new sequential scale with a logarithmic
    transform, analogous to a log scale.

    Parameters
    ----------
    domain : list[int | float]
        Domain
    interpolator : Callable
        Interpolator

    Returns
    -------
    SequentialLog
        Sequential object

    Examples
    --------

    >>> d3.scale_sequential_log([0, 100], d3.interpolate_blues)
    """
    scale = SequentialLog()
    if len(args) == 1:
        return init_interpolator(scale, interpolator=args[0])
    elif len(args) == 2:
        domain, interpolator = args
        return init_interpolator(scale, domain=domain, interpolator=interpolator)
    return init_interpolator(scale)


@overload
def scale_sequential_symlog() -> SequentialSymlog: ...


@overload
def scale_sequential_symlog(interpolator: Callable) -> SequentialSymlog: ...


@overload
def scale_sequential_symlog(
    domain: list[int | float], interpolator: Callable
) -> SequentialSymlog: ...


def scale_sequential_symlog(*args):
    """
    Builds a new sequential scale with a symmetric
    logarithmic transform, analogous to a symlog scale.

    Parameters
    ----------
    domain : list[int | float]
        Domain
    interpolator : Callable
        Interpolator

    Returns
    -------
    SequentialSymlog
        Sequential object

    Examples
    --------

    >>> d3.scale_sequential_symlog([0, 100], d3.interpolate_blues)
    """
    scale = SequentialSymlog()
    if len(args) == 1:
        return init_interpolator(scale, interpolator=args[0])
    elif len(args) == 2:
        domain, interpolator = args
        return init_interpolator(scale, domain=domain, interpolator=interpolator)
    return init_interpolator(scale)


@overload
def scale_sequential_pow() -> SequentialPow: ...


@overload
def scale_sequential_pow(interpolator: Callable) -> SequentialPow: ...


@overload
def scale_sequential_pow(
    domain: list[int | float], interpolator: Callable
) -> SequentialPow: ...


def scale_sequential_pow(*args):
    """
    Builds a new sequential scale with an exponential
    transform, analogous to a power scale.


    Parameters
    ----------
    domain : list[int | float]
        Domain
    interpolator : Callable
        Interpolator

    Returns
    -------
    SequentialPow
        Sequential object

    Examples
    --------

    >>> d3.scale_sequential_pow([0, 100], d3.interpolate_blues)
    """
    scale = SequentialPow()
    if len(args) == 1:
        return init_interpolator(scale, interpolator=args[0])
    elif len(args) == 2:
        domain, interpolator = args
        return init_interpolator(scale, domain=domain, interpolator=interpolator)
    return init_interpolator(scale)


@overload
def scale_sequential_sqrt() -> SequentialPow: ...


@overload
def scale_sequential_sqrt(interpolator: Callable) -> SequentialPow: ...


@overload
def scale_sequential_sqrt(
    domain: list[int | float], interpolator: Callable
) -> SequentialPow: ...


def scale_sequential_sqrt(*args):
    """
    Builds a new sequential scale with a square-root
    transform, analogous to a sqrt scale


    Parameters
    ----------
    domain : list[int | float]
        Domain
    interpolator : Callable
        Interpolator

    Returns
    -------
    SequentialSqrt
        Sequential object

    Examples
    --------

    >>> d3.scale_sequential_sqrt([0, 100], d3.interpolate_blues)
    """
    return scale_sequential_pow(*args).set_exponent(0.5)



from __future__ import annotations

from collections.abc import Callable, Iterator
from itertools import zip_longest
from typing import Any, Protocol, TypeAlias, overload

from lxml import etree

from .attr import attr_constant, attr_function
from .bind import bind_index, bind_key
from .clone import clone
from .constant import constant
from .enter import EnterNode
from .namespace import namespace
from .style import style_constant, style_function, style_value
from .text import text_constant, text_function

Data: TypeAlias = Any
Value: TypeAlias = Any


class Accessor(Protocol):
    @overload
    def __call__(self, d: Data) -> Value: ...

    @overload
    def __call__(self, d: Data, i: int) -> Value: ...

    @overload
    def __call__(self, d: Data, i: int, group: list[etree.Element]) -> Value: ...

    def __call__(self, *args) -> Value: ...


def selector(element: etree.Element, selection: str | None = None):
    if selection is None:
        return element
    order = ""
    if ":" in selection:
        selection, order = selection.split(":")
        if order != "last-of-type":
            raise ValueError("Only 'last-of-type' is implemented currently.")
        order = "[last()]"
    if "." in selection:
        tag, class_name = selection.split(".")
        tag = tag or "*"
        class_name = f"[@class='{class_name}']" if class_name else ""
        return element.xpath(f"./*/{tag}{order}{class_name}") + element.xpath(
            f"./{tag}{order}{class_name}"
        )
    return element.xpath(f"./*/{selection}{order}") + element.xpath(
        f"./{selection}{order}"
    )


def creator(node: etree.Element, fullname: dict | None = None):
    return (
        etree.SubElement(node, fullname["local"], attrs=fullname["space"])
        if isinstance(fullname, dict)
        else etree.SubElement(node, fullname)
    )


class Selection:
    """
    A selection is a set of elements from the DOM.
    Typically these elements are identified by selectors such
    as .fancy for elements with the class fancy, or div to
    select DIV elements.

    Selection methods come in two forms, :code:`select` and :code:`select_all`:
    the former selects only the first matching element, while the latter selects
    all matching elements in document order. The top-level selection methods,
    d3.select and d3.select_all, query the entire document; the subselection
    methods, selection.select and selection.select_all, restrict selection to
    descendants of the selected elements.

    By convention, selection methods that return the current selection such
    as selection.attr use four spaces of indent, while methods that return
    a new selection use only two. This helps reveal changes of context by
    making them stick out of the chain:

    Parameters
    ----------
    groups : list[list[etree.Element]]
        List of groups of selected nodes given its parent.
    parents : list[etree.Element]
        List of parents related to groups.
    enter : list[EnterNode] | None = None
        List of placeholder nodes for each datum that had no corresponding
        DOM element in the selection.
    exit : list[etree.Element] = None
        List of existing DOM elements in the selection for which no new datum was found.
    data : dict[etree.Element, Data] | None = None
        Association between nodes and its data

    Examples
    --------

    >>> (
    ...     d3.create("body")
    ...     .append("svg")
    ...     .attr("width", 960)
    ...     .attr("height", 500)
    ...     .append("g")
    ...     .attr("transform", "translate(20, 20)")
    ...     .append("rect")
    ...     .attr("width", 920)
    ...     .attr("height", 460)
    ... )
    """

    def __init__(
        self,
        groups: list[list[etree.Element]],
        parents: list[etree.Element],
        enter: list[EnterNode] | None = None,
        exit: list[etree.Element] = None,
        data: dict[etree.Element, Data] | None = None,
    ):
        self._groups = groups
        self._parents = parents
        self._enter = enter
        self._exit = exit
        self._data = data or {}

    def select(self, selection: str | None = None) -> Selection:
        """
        Selects the first element that matches the specified :code:`selection` string

        Parameters
        ----------
        selection : str | None
            Selection string

        Returns
        -------
        Selection
            Selection of first element

        Examples
        --------

        >>> d3.select("g.ticks")

        Notes
        -----

        Supported strings are :code:`<tag_name>.<class_name>` or `.<class_name>` or `<tag_name>`.
        To get element in reverse order, you can use `<tag_name>:last-of-type`.
        """
        subgroups = [
            selector(node, selection)[:1]
            for group in self._groups
            for node in group
            if node is not None
        ]
        parents = [
            (group[0]._parent if isinstance(group[0], EnterNode) else group[0])
            for group in self._groups
            if group[0] is not None
        ]
        return Selection(subgroups, parents or self._parents, data=self._data)

    def select_all(self, selection: str | None = None) -> Selection:
        """
        Selects all elements that match the specified :code:`selection` string

        Parameters
        ----------
        selection : str | None
            Selection string

        Returns
        -------
        Selection
            Selection of all matched elements

        Examples
        --------

        >>> d3.select_all("g.ticks")

        Notes
        -----

        Supported strings are :code:`<tag_name>.<class_name>` or `.<class_name>` or `<tag_name>`.
        To get element in reverse order, you can use `<tag_name>:last-of-type`.
        """
        subgroups = [
            selector(node, selection)
            for group in self._groups
            for node in group
            if node is not None
        ]
        parents = [
            (group[0]._parent if isinstance(group[0], EnterNode) else group[0])
            for group in self._groups
            if group[0] is not None
        ]
        return Selection(subgroups, parents or self._parents, data=self._data)

    def enter(self) -> Selection:
        """
        Returns the enter selection: placeholder nodes for each datum
        that had no corresponding DOM element in the selection.

        Returns
        -------
        Selection
            Enter selection
        """
        return Selection(
            self._enter or [[None] * len(group) for group in self._groups],
            self._parents,
            data=self._data,
        )

    def exit(self) -> Selection:
        """
        Returns the exit selection: existing DOM elements in the
        selection for which no new datum was found.
        (The exit selection is empty for selections not returned
        by selection.data.)


        Returns
        -------
        Selection
            Exit selection
        """
        return Selection(
            self._exit or [[None] * len(group) for group in self._groups],
            self._parents,
            data=self._data,
        )

    def merge(self, context: Selection) -> Selection:
        """
        Returns a new selection merging this selection with the
        specified other selection or transition. The returned
        selection has the same number of groups and the same parents
        as this selection. Any missing (None) elements in this
        selection are filled with the corresponding element, if
        present (not null), from the specified selection. (If the
        other selection has additional groups or parents, they are
        ignored.)

        Parameters
        ----------
        context : Selection
            Selection

        Returns
        -------
        Selection
            Merged selection
        """
        selection = context.selection() if hasattr(context, "selection") else context

        merges = []
        for groups0, groups1 in zip_longest(
            self._groups, selection._groups, fillvalue=[]
        ):
            merge = []
            for group0, group1 in zip_longest(groups0, groups1, fillvalue=None):
                node = group0 if group0 is not None else group1
                merge.append(node)
            merges.append(merge)

        for j in range(len(merges), len(self._groups)):
            merges.append(self._groups[j])

        return Selection(merges, self._parents, data=self._data | selection._data)

    def append(self, name: str) -> Selection:
        """
        If the specified name is a string, appends a new element
        of this type (tag name) as the last child of each selected
        element, or before the next following sibling in the update
        selection if this is an enter selection. The latter behavior
        for enter selections allows you to insert elements into the
        DOM in an order consistent with the new bound data; however,
        note that selection.order may still be required if updating
        elements change order (i.e., if the order of new data is
        inconsistent with old data).

        Parameters
        ----------
        name : str
            Tag name

        Returns
        -------
        Selection
            Selection

        Examples
        --------

        >>> svg.select_all("g").append("path")
        """
        fullname = namespace(name)
        subgroups = []
        for group in self._groups:
            subgroup = []
            for node in group:
                if node is None:
                    continue
                if isinstance(node, EnterNode):
                    enter_node = node
                    node = enter_node._parent
                    subnode = creator(node, fullname)
                    node.append(subnode)
                    subgroup.append(subnode)
                    self._data[subnode] = enter_node.__data__
                else:
                    subnode = creator(node, fullname)
                    node.append(subnode)
                    subgroup.append(subnode)
                    self._data[subnode] = self._data.get(node)
            subgroups.append(subgroup)
        parents = [
            (group[0]._parent if isinstance(group[0], EnterNode) else group[0])
            for group in self._groups
            if group[0] is not None
        ]
        return Selection(subgroups, parents or self._parents, data=self._data)

    def each(
        self, callback: Callable[[etree.Element, Data, int, list[etree.Element]], None]
    ):
        """
        Invokes the specified function for each selected element,
        in order, being passed the current DOM element (nodes[i]),
        the current datum (d), the current index (i), and the
        current group (nodes).

        Parameters
        ----------
        callback : Callable[[etree.Element, Data, int, list[etree.Element]], None]
            Function to call which takes as argument: node, data, index and group
        """
        for group in self._groups:
            for i, node in enumerate(group):
                if node is not None:
                    if isinstance(node, EnterNode):
                        node = node._parent
                    callback(node, self._data.get(node), i, group)

    def attr(self, name: str, value: Accessor | str | None = None) -> Selection:
        """
        If a value is specified, sets the attribute with the specified name
        to the specified value on the selected elements and returns this selection.

        If the value is a function, it is evaluated for each selected element,
        in order, being passed the current datum (d), the current index (i),
        and the current group (nodes).

        Parameters
        ----------
        name : str
            Name of the attribute
        value : Accessor | str | None
            Value

        Returns
        -------
        Selection
            Itself

        Examples
        --------

        >>> selection.attr("color", "red")
        """
        if value is None:
            return self.node().get(name)
        elif callable(value):
            self.each(attr_function(name, value))
        else:
            self.each(attr_constant(name, value))
        return self

    def style(self, name: str, value: Accessor | str | None = None) -> Selection:
        """
        If a value is specified, sets the style with the specified name
        to the specified value on the selected elements and returns this selection.

        If the value is a function, it is evaluated for each selected element,
        in order, being passed the current datum (d), the current index (i),
        and the current group (nodes).

        Parameters
        ----------
        name : str
            Name of the style
        value : Accessor | str | None
            Value constant or function

        Returns
        -------
        Selection
            Itself

        Examples
        --------

        >>> selection.style("color", "red")
        """
        if value is None:
            return style_value(self.nodes.get(name).get("style"), name)
        elif callable(value):
            self.each(style_function(name, value))
        else:
            self.each(style_constant(name, value))
        return self

    def text(self, value: Accessor | str | None = None) -> Selection:
        """
        If the value is a constant, then all elements are given the same
        text content; otherwise, if the value is a function, it is evaluated
        for each selected element, in order, being passed the current
        datum (d), the current index (i), and the current group (nodes).

        Parameters
        ----------
        value : Accessor | str | None
            Value constant or function

        Returns
        -------
        Selection
            Itself

        Examples
        --------

        >>> selection.text("Hello, world!")
        """
        if value is None:
            return self.node().get("text")
        elif callable(value):
            self.each(text_function(value))
        else:
            self.each(text_constant(value))
        return self

    def datum(self, value: Data) -> Selection:
        """
        Sets the bound data for the first selected node.

        Parameters
        ----------
        value : Data
            Value

        Returns
        -------
        Selection
            Itself
        """
        self._data[self.node()] = value
        return self

    def data(self, values: list[Data], key: Accessor | None = None) -> Selection:
        """
        Binds the specified list of data with the selected elements,
        returning a new selection that represents the update selection:
        the elements successfully bound to data. Also defines the enter
        and exit selections on the returned selection, which can be used
        to add or remove elements to correspond to the new data. The
        specified data is an array of arbitrary values (e.g., numbers or
        objects), or a function that returns an array of values for each
        group.

        Parameters
        ----------
        values : list[Data]
            List of data to bind
        key : Accessor | None
            Optional accessor which returns a key value

        Returns
        -------
        Selection
            Itself
        """
        bind = bind_key if key else bind_index
        parents = self._parents
        groups = self._groups

        if not callable(values):
            values = constant(values)

        update = [None] * len(groups)
        enter = [None] * len(groups)
        exit = [None] * len(groups)
        for j in range(len(groups)):
            parent = parents[j]
            group = groups[j]
            data = list(values(parent, self._data.get(parent), j, parents))
            enter[j] = enter_group = [None] * len(data)
            update[j] = update_group = [None] * len(data)
            exit[j] = exit_group = [None] * len(group)

            bind(
                self._data,
                parent,
                group,
                enter_group,
                update_group,
                exit_group,
                data,
                key,
            )

            i1 = 0
            for i0 in range(len(data)):
                if previous := enter_group[i0]:
                    if i0 >= i1:
                        i1 = i0 + 1
                    while not (
                        i1 < len(update_group) and update_group[i1] is not None
                    ) and i1 < len(data):
                        i1 += 1
                    previous._next = update_group[i1] if i1 < len(data) else None

        return Selection(update, parents, enter, exit, self._data)

    def order(self) -> Selection:
        """
        Re-inserts elements into the document such that the document order
        of each group matches the selection order.

        Returns
        -------
        Selection
            Itself
        """
        for group in self._groups:
            next_node = None
            for node in reversed(group):
                if node is not None:
                    if next_node is not None and (node.getprevious() == next_node):
                        parent = next_node.getparent()
                        index = parent.index(next_node)
                        parent.pop(index)
                        parent.insert(index, node)
                    next_node = node
        return self

    def join(
        self,
        onenter: Callable | Selection,
        onupdate: Callable | None = None,
        onexit: Callable | None = None,
    ):
        """
        Appends, removes and reorders elements as necessary to match
        the data that was previously bound by selection.data, returning
        the merged enter and update selection. This method is a convenient
        alternative to the explicit general update pattern, replacing
        selection.enter, selection.exit, selection.append, selection.remove, and selection.order.

        Parameters
        ----------
        onenter : Callable | Selection
            Enter selection or function
        onupdate : Callable | None
            Function
        onexit : Callable | None
            Function

        Returns
        -------
            Selection with joined elements
        """
        enter = self.enter()
        update = self
        exit = self.exit()

        # Enter
        if callable(onenter):
            enter = onenter(enter)
            if enter:
                enter = enter.selection()
        else:
            enter = enter.append(onenter)

        # Update
        if onupdate is not None:
            update = onupdate(update)
            if update:
                update = update.selection()

        # Exit
        if onexit is None:
            exit.remove()
        else:
            onexit(exit)

        if enter and update:
            return enter.merge(update).order()
        else:
            return update

    def insert(self, name: str, before: etree.Element) -> Selection:
        """
        If the specified name is a string, inserts a new element
        of this type (tag name) before the first element matching
        the specified before selector for each selected element.

        Parameters
        ----------
        name : str
            Tag name
        before : etree.Element
            Node element

        Returns
        -------
        Selection
            Itself with inserted element
        """
        fullname = namespace(name)
        for group in self._groups:
            for i, node in enumerate(group):
                if node is not None:
                    if isinstance(node, EnterNode):
                        node = node._parent
                    selection = selector(node, before)
                    if selection := [
                        found for found in selection if found.getparent() == node
                    ]:
                        index = node.index(selection[0])
                        created = creator(node, fullname)
                        node.insert(index, created)
                        group[i] = created
        return self

    def remove(self) -> Selection:
        """
        Removes the selected elements from the document. Returns this
        selection (the removed elements) which are now detached from
        the DOM.

        Returns
        -------
        Selection
            Itself with removed elements
        """

        def remove(node, data, i, group):
            parent = node.getparent()
            if parent is not None:
                parent.remove(node)

        self.each(remove)
        return self

    def call(self, func: Callable[[Selection, ...], Any], *args: Data) -> Selection:
        """
        Invokes the specified function exactly once, passing in
        this selection along with any optional arguments. Returns
        this selection.

        Parameters
        ----------
        func : Callable[[Selection, ...], Any]
            Function to call
        args : Data
            Arguments for the function to call

        Returns
        -------
        Selection
            Itself

        Examples
        --------

        This is equivalent to invoking the function
        by hand but facilitates method chaining. For example, to set
        several styles in a reusable function:

        >>> def name(selection, first, last):
        ...     selection.attr("first-name", first).attr("last-name", last)

        Now say:
        >>> d3.select_all("div").call(name, "John", "Snow")

        This is roughly equivalent to:

        >>> name(d3.select_all("div"), "John", "Snow")
        """
        func(self, *args)
        return self

    def clone(self) -> Selection:
        """
        Inserts clones of the selected elements immediately following
        the selected elements and returns a selection of the newly
        added clones.

        Returns
        -------
        Selection
            Clone of itself
        """
        subgroups = [
            clone(node) for group in self._groups for node in group if node is not None
        ]
        return Selection(subgroups, self._parents, data=self._data)

    def node(self) -> etree.Element:
        """
        Returns the first (non-None) element in this selection.

        Returns
        -------
        etree.Element
            Node
        """
        return next(iter(self))

    def nodes(self) -> list[etree.Element]:
        """
        Returns a list of all (non-None) elements in this selection.

        Returns
        -------
        list[etree.Element]
            List of nodes
        """
        return list(self)

    def __iter__(self) -> Iterator[etree.Element]:
        """
        Make the selection as an iterator

        Returns
        -------
        Iterator[etree.Element]
            Iterator of non-None nodes
        """
        for group in self._groups:
            for node in group:
                if node is not None:
                    yield node

    def selection(self) -> Selection:
        """
        Returns the selection

        Returns
        -------
        Selection
            Itself
        """
        return self

    def to_string(self, pretty_print: bool = True) -> str:
        """
        Convert selection to string

        Parameters
        ----------
        pretty_print : bool
            :code:`True` to prettify output

        Returns
        -------
        str
            String
        """
        return etree.tostring(self._parents[0], pretty_print=pretty_print).decode(
            "utf-8"
        )

    def __str__(self) -> str:
        """
        Returns the SVG content

        Returns
        -------
        str
            String
        """
        return self.to_string(False)

    def __repr__(self) -> str:
        """
        Represents the selection

        Returns
        -------
        str
            String
        """

        def node_repr(node):
            if node is None:
                return str(node)
            if isinstance(node, EnterNode):
                return str(node)
            tag = node.tag
            class_name = node.attrib.get("class")
            if class_name:
                return f"{tag}.{class_name}"
            return tag

        groups = [
            f"[{', '.join(node_repr(node) for node in group)}]"
            for group in self._groups
        ]
        parents = f"[{', '.join(node_repr(parent) for parent in self._parents)}]"
        return (
            f"Selection(\n    groups=[{', '.join(groups)}],\n    parents={parents},"
            f"\n    enter={self._enter},\n    exit={self._exit},\n    data={self._data},\n)"
        )



from ...interpolate import interpolate_cubehelix_long
from ...coloration import cubehelix
from math import floor

interpolate_warm = interpolate_cubehelix_long(cubehelix(-100, 0.75, 0.35), cubehelix(80, 1.50, 0.8))
interpolate_cool = interpolate_cubehelix_long(cubehelix(260, 0.75, 0.35), cubehelix(80, 1.50, 0.8))


def interpolate_rainbow(t):
    if t < 0 or t > 1:
        t -= floor(t)
    ts = abs(t - 0.5)
    c = cubehelix(360 * t - 100, 1.5 - 1.5 * ts, 0.8 - 0.9 * ts)
    return str(c)



from __future__ import annotations

import math
from bisect import bisect
from collections.abc import Callable
from datetime import datetime
from typing import Any, Generic, TypeVar

from ..interpolate import (
    interpolate as interpolate_value,
)
from ..interpolate import (
    interpolate_number,
    interpolate_round,
)
from .constant import constant
from .number import number

T = TypeVar("T")


def identity(x):
    return x


def normalize(a, b):
    if isinstance(a, datetime):
        a = a.timestamp()
    elif isinstance(a, str):
        a = float(a)
    if isinstance(b, datetime):
        b = b.timestamp()
    elif isinstance(b, str):
        b = float(b)
    b = b - a
    if not math.isnan(b) and b:

        def f(x):
            if isinstance(x, datetime):
                x = x.timestamp()
            elif isinstance(x, str):
                x = float(x)
            return (x - a) / b

        return f
    else:
        return constant(math.nan if math.isnan(b) else 0.5)


def clamper(a, b):
    a = a.timestamp() if isinstance(a, datetime) else a
    b = b.timestamp() if isinstance(b, datetime) else b
    if a > b:
        a, b = b, a

    def f(x):
        x = x.timestamp() if isinstance(x, datetime) else x
        return max(a, min(b, x))

    return f


class BiMap:
    def __init__(self, domain, range_vals, interpolate):
        d0, d1 = domain[0], domain[1]
        r0, r1 = range_vals[0], range_vals[1]
        if d1 < d0:
            self.d0 = normalize(d1, d0)
            self.r0 = interpolate(r1, r0)
        else:
            self.d0 = normalize(d0, d1)
            self.r0 = interpolate(r0, r1)

    def __call__(self, x):
        return self.r0(self.d0(x))


class PolyMap:
    def __init__(self, domain, range_vals, interpolate):
        self.domain = domain
        self.j = j = min(len(self.domain), len(range_vals)) - 1
        self.d = [None] * j
        self.r = [None] * j

        if self.domain[j] < self.domain[0]:
            self.domain = self.domain[::-1]
            range_vals = range_vals[::-1]

        for i in range(j):
            self.d[i] = normalize(self.domain[i], self.domain[i + 1])
            self.r[i] = interpolate(range_vals[i], range_vals[i + 1])

    def __call__(self, x):
        i = bisect(self.domain, x, 1, self.j) - 1
        return self.r[i](self.d[i](x))


def copy(source, target):
    return (
        target.set_domain(source.domain)
        .set_range(source.range)
        .set_interpolate(source.interpolate)
        .set_clamp(source.clamp)
        .set_unknown(source.unknown)
    )


class Transformer(Generic[T]):
    """
    Continous transformation

    Parameters
    ----------
    t : Callable[[int | float], T]
        Transform function
    u : Callable[[T], int | float]
        Untransform function
    """

    def __init__(
        self,
        t: Callable[[int | float], T] = identity,
        u: Callable[[T], int | float] = identity,
    ):
        self.transform = t
        self.untransform = u
        self._domain = [0, 1]
        self._range = [0, 1]
        self._clamp = identity
        self._interpolate = interpolate_value
        self._unknown = None
        self.input = None
        self.output = None
        self.rescale()

    def rescale(self):
        n = min(len(self._domain), len(self._range))
        if self._clamp != identity:
            self._clamp = clamper(self._domain[0], self._domain[n - 1])
        self.piecewise = PolyMap if n > 2 else BiMap
        self.output = self.input = None
        return self

    def __call__(self, x: int | float) -> T:
        """
        Given a value from the domain, returns the corresponding value from the range.

        Parameters
        ----------
        x : int | float
            Input value

        Returns
        -------
        T
            Corresponding value from the range
        """
        if x is None or (isinstance(x, float) and math.isnan(x)):
            return self._unknown
        else:
            if not self.output:
                domain = [
                    x.timestamp() if isinstance(x, datetime) else x
                    for x in self._domain
                ]
                domain = [self.transform(x) for x in domain]
                self.output = self.piecewise(domain, self._range, self._interpolate)
            return self.output(self.transform(self._clamp(x)))

    def invert(self, y: T) -> int | float:
        """
        Given a value from the range, returns the corresponding value
        from the domain. Inversion is useful for interaction, say to
        determine the data value corresponding to the position of the mouse.

        Parameters
        ----------
        y : T
            Input value

        Returns
        -------
        int | float
            Corresponding value from the domain
        """
        if not self.input:
            domain = [
                x.timestamp() if isinstance(x, datetime) else x for x in self._domain
            ]
            domain = [self.transform(x) for x in domain]
            self.input = self.piecewise(self._range, domain, interpolate_number)
        return self._clamp(self.untransform(self.input(y)))

    def set_domain(self, domain: list[int | float]) -> Transformer:
        """
        Sets the scale's domain to the specified array of numbers

        Parameters
        ----------
        domain : list[int | float]
            Domain

        Returns
        -------
        Transformer
            Itself
        """
        self._domain = list(map(number, domain))
        return self.rescale()

    @property
    def domain(self) -> list[int | float]:
        return self._domain.copy()

    def set_range(self, range_vals: list[T]) -> Transformer:
        """
        Sets the scale's range to the specified array of values

        Parameters
        ----------
        range_vals : list[T]
            Range values

        Returns
        -------
        Transformer
            Itself
        """
        self._range = list(range_vals)
        return self.rescale()

    @property
    def range(self) -> list[T]:
        return self._range.copy()

    def set_range_round(self, range_vals: list[T]) -> Transformer:
        """
        Sets the scale's range to the specified array of values
        and sets scale's interpolator to :code:`interpolate_round`.

        Parameters
        ----------
        range_vals : list[T]
            Range values

        Returns
        -------
        Transformer
            Itself
        """
        self._range = list(map(float, range_vals))
        self._interpolate = interpolate_round
        return self.rescale()

    def set_clamp(self, clamp: bool) -> Transformer:
        """
        Enables or disables clamping accordingly

        Parameters
        ----------
        clamp : bool
            Clamp value

        Returns
        -------
        Transformer
            Itself
        """
        self._clamp = True if clamp else identity
        return self.rescale()

    @property
    def clamp(self) -> bool:
        return self._clamp != identity

    def set_interpolate(self, interpolate: Callable[[T, T], T]) -> Transformer:
        """
        Sets the scale's range interpolator factory.

        Parameters
        ----------
        interpolate : Callable[[T, T], T]
            Interpolate function

        Returns
        -------
        Transformer
            Itself
        """
        self._interpolate = interpolate
        return self.rescale()

    @property
    def interpolate(self) -> Callable[[T, T], T]:
        return self._interpolate

    def set_unknown(self, unknown: Any) -> Transformer:
        """
        Sets the output value of the scale for undefined
        or NaN input values.

        Parameters
        ----------
        unknown : Any
            Unknown value

        Returns
        -------
        Transformer
            Itself
        """
        self._unknown = unknown
        return self.rescale()

    @property
    def unknown(self) -> Any:
        return self._unknown



from .object_to_value import interpolate as interpolate_value


def piecewise(interpolate, values=None):
    if values is None:
        values = interpolate
        interpolate = interpolate_value

    n = len(values) - 1
    I = [interpolate_value(values[i], values[i + 1]) for i in range(n)]

    def local_interpolate(t):
        i = max(0, min(n - 1, int(t * n)))
        return I[i](t * n - i)

    return local_interpolate



from ..time import (
    utcDay,
    utcHour,
    utcMinute,
    utcMonth,
    utcSecond,
    utcTickInterval,
    utcTicks,
    utcWeek,
    utcYear,
)
from ..time_format import utcFormat
from .init import init_range
from .time import calendar


def utc_time():
    return init_range(
        calendar(
            utcTicks,
            utcTickInterval,
            utcYear,
            utcMonth,
            utcWeek,
            utcDay,
            utcHour,
            utcMinute,
            utcSecond,
            utcFormat,
        ).domain([datetime.datetime(2000, 1, 1), datetime.datetime(2000, 1, 2)])
    )



from __future__ import annotations

import math
from collections.abc import Callable
from typing import TypeVar, overload

from .continuous import Transformer, copy, identity
from .init import init_range
from .linear import LinearBase

T = TypeVar("T")


def transform_pow(exponent):
    return lambda x: (-math.pow(-x, exponent) if x < 0 else math.pow(x, exponent))


def transform_sqrt(x):
    return -math.sqrt(-x) if x < 0 else math.sqrt(x)


def transform_square(x):
    return -x * x if x < 0 else x * x


class ScalePow(Transformer, LinearBase):
    """
    Power ("pow") scales are similar to linear scales, except an exponential
    transform is applied to the input domain value before the output range
    value is computed.
    Each range value y can be expressed as a function of the domain value x:
    :math:`y = m \\cdot x^k + b`, where :math:`k` is the exponent value.
    Power scales also support negative domain values, in which case the
    input value and the resulting output value are multiplied by -1.

    Parameters
    ----------
    t : Callable
        Transform function
    u : Callable
        Untransform function
    exponent : float | int
        Exponent
    """

    def __init__(
        self, t: Callable = identity, u: Callable = identity, exponent: float | int = 1
    ):
        super().__init__(t, u)
        self._exponent = exponent

    def _rescale(self):
        if self._exponent == 1:
            self.transform = identity
            self.untransform = identity
            self.rescale()
            return self
        elif self._exponent == 0.5:
            self.transform = transform_sqrt
            self.untransform = transform_square
            self.rescale()
            return self
        else:
            self.transform = transform_pow(self._exponent)
            self.untransform = transform_pow(1 / self._exponent)
            self.rescale()
            return self

    def set_exponent(self, exponent: int | float) -> ScalePow:
        """
        Sets the current exponent to the given numeric value.

        Parameters
        ----------
        exponent : int | float
            exponent

        Returns
        -------
        ScalePow
            Itself
        """
        self._exponent = float(exponent)
        return self._rescale()

    @property
    def exponent(self):
        return self._exponent

    def copy(self):
        return copy(self, ScalePow()).set_exponent(self.exponent)


@overload
def scale_pow() -> ScalePow: ...


@overload
def scale_pow(range_vals: list[T]) -> ScalePow: ...


@overload
def scale_pow(domain: list[int | float], range_vals: list[T]) -> ScalePow: ...


def scale_pow(*args) -> ScalePow:
    """
    Constructs a new pow scale with the specified domain and
    range, the exponent 1, the default interpolator and
    clamping disabled.

    Returns
    -------
    ScalePow
        Scale object
    """
    scale = ScalePow()
    if len(args) == 1:
        return init_range(scale, range_vals=args[0])
    elif len(args) == 2:
        domain, range_vals = args
        return init_range(scale, domain=domain, range_vals=range_vals)
    return init_range(scale)


@overload
def scale_sqrt() -> ScalePow: ...


@overload
def scale_sqrt(range_vals: list[T]) -> ScalePow: ...


@overload
def scale_sqrt(domain: list[int | float], range_vals: list[T]) -> ScalePow: ...


def scale_sqrt(*args) -> ScalePow:
    """
    Constructs a new pow scale with the specified domain and range,
    the exponent 0.5, the default interpolator and clamping disabled.

    Returns
    -------
    ScalePow
        Scale object
    """
    scale = ScalePow().set_exponent(0.5)
    if len(args) == 1:
        return init_range(scale, range_vals=args[0])
    elif len(args) == 2:
        domain, range_vals = args
        return init_range(scale, domain=domain, range_vals=range_vals)
    return init_range(scale)



from collections.abc import Callable
from math import cosh, dist, exp, log, sinh, sqrt, tanh

EPSILON = 1e-12
SQRT2 = sqrt(2)


class Base:
    def __init__(self, rho, p0, p1):
        ux0, uy0, w0 = p0
        ux1, uy1, w1 = p1
        dx = ux1 - ux0
        dy = uy1 - uy0
        self.rho = rho
        self.ux0 = ux0
        self.uy0 = uy0
        self.w0 = w0
        self.w1 = w1
        self.dx = dx
        self.dy = dy
        self.d2 = dx * dx + dy * dy


class SingularZoomRhoInterpolator(Base):
    def __init__(self, rho, rho2, rho4, p0, p1):
        super().__init__(rho, p0, p1)
        w0 = self.w0
        w1 = self.w1
        self.s = log(w1 / w0) / rho
        self.duration = self.s * 1000 * rho / SQRT2

    def __call__(self, t):
        return [
            self.ux0 + t * self.dx,
            self.uy0 + t * self.dy,
            self.w0 * exp(self.rho * t * self.s),
        ]


class GeneralZoomRhoInterpolator(Base):
    def __init__(self, rho, rho2, rho4, p0, p1):
        super().__init__(rho, p0, p1)
        d2 = self.d2
        w0 = self.w0
        w1 = self.w1

        self.rho2 = rho2
        self.d1 = d1 = sqrt(d2)

        b0 = (w1 * w1 - w0 * w0 + rho4 * d2) / (2 * w0 * rho2 * d1)
        b1 = (w1 * w1 - w0 * w0 - rho4 * d2) / (2 * w1 * rho2 * d1)
        self.r0 = r0 = log(sqrt(b0 * b0 + 1) - b0)
        self.r1 = r1 = log(sqrt(b1 * b1 + 1) - b1)

        self.s = (r1 - r0) / rho
        self.duration = self.s * 1000 * self.rho / SQRT2

    def __call__(self, t):
        c = self.w0 / (self.rho2 * self.d1)
        u = c * (cosh(self.r0) * tanh(self.rho * t * self.s + self.r0) - sinh(self.r0))
        return [
            self.ux0 + u * self.dx,
            self.uy0 + u * self.dy,
            self.w0 * cosh(self.r0) / cosh(self.rho * t * self.s + self.r0),
        ]


class ZoomRho:
    """
    An interpolator for zooming smoothly between two views of a two-dimensional
    plane based on “Smooth and efficient zooming and panning” by
    Jarke J. van Wijk and Wim A.A. Nuij.
    """

    def __init__(self, rho, rho2, rho4):
        self.rho = rho
        self.rho2 = rho2
        self.rho4 = rho4

    def __call__(self, a: list, b: list) -> Callable:
        """
        Returns an interpolator between the two views a and b. Each view is defined
        as an array of three numbers: cx, cy and width. The first two coordinates cx,
        cy represent the center of the viewport; the last coordinate width represents
        the size of the viewport.

        Parameters
        ----------
        a : list
            View a
        b : list
            View b

        Returns
        -------
        Callable
            Interpolator
        """
        d = dist(a[:-1], b[:-1]) ** 2
        interpolator = (
            SingularZoomRhoInterpolator if d < EPSILON else GeneralZoomRhoInterpolator
        )
        return interpolator(self.rho, self.rho2, self.rho4, a, b)

    def set_rho(self, new_rho):
        new_rho = max(1e-3, float(new_rho))
        return ZoomRho(new_rho, new_rho * new_rho, new_rho**4)


interpolate_zoom = ZoomRho(SQRT2, 2, 4)



class EnterNode:
    def __init__(self, parent, datum):
        self._next = None
        self._parent = parent
        self.__data__ = datum

    def __str__(self):
        if self._parent is None:
            return f"EnterNode({self._parent}, {self.__data__})"
        tag = self._parent.tag
        class_name = self._parent.attrib.get("class")
        if class_name:
            return f"EnterNode({tag}.{class_name}, {self.__data__})"
        return f"EnterNode({tag}, {self.__data__})"

    def __repr__(self):
        return str(self)

    # def append_child(self, child):
    #     return self._parent.insert_before(child, self._next)
    #
    # def insert_before(self, child, next):
    #     return self._parent.insert_before(child, next)
    #
    # def query_selector(self, selector):
    #     return self._parent.query_selector(selector)
    #
    # def query_selector_all(self, selector):
    #     return self._parent.query_selector_all(selector)



from __future__ import annotations

from typing import TypeVar, overload

from .init import init_range
from .ordinal import ScaleOrdinal

T = TypeVar("T")


class ScaleBand(ScaleOrdinal):
    """
    Band scales are like band scales except the output range is continuous
    and numeric. The scale divides the continuous range into uniform bands.
    Band scales are typically used for bar charts with an band or categorical
    dimension.
    """

    def __init__(self):
        super().__init__()
        self._r0 = 0
        self._r1 = 1
        self._step = None
        self._bandwidth = None
        self._round = False
        self._padding_inner = 0
        self._padding_outer = 0
        self._align = 0.5
        self.rescale()

    def rescale(self):
        n = len(self._domain)
        reverse = self._r1 < self._r0
        start = self._r1 if reverse else self._r0
        stop = self._r0 if reverse else self._r1
        self._step = (stop - start) / max(
            1, n - self._padding_inner + self._padding_outer * 2
        )
        if self._round:
            self._step = int(self._step)
        start += (stop - start - self._step * (n - self._padding_inner)) * self._align
        self._bandwidth = self._step * (1 - self._padding_inner)
        if self._round:
            start = round(start)
            self._bandwidth = round(self._bandwidth)
        values = [start + self._step * i for i in range(n)]
        return super().set_range(values[::-1] if reverse else values)

    def set_domain(self, domain: list[T]) -> ScaleBand:
        """
        Sets the scale's domain to the specified array of values

        Parameters
        ----------
        domain : list[T]
            Domain

        Returns
        -------
        ScaleBand
            Itself
        """
        super().set_domain(domain)
        return self.rescale()

    @property
    def domain(self):
        return self._domain.copy()

    def set_range(self, range_vals: list[int | float]) -> ScaleBand:
        """
        Sets the scale's range to the specified array of numbers

        Parameters
        ----------
        range_vals : list[int | float]
            Range

        Returns
        -------
        ScaleBand
            Itself
        """
        self._r0, self._r1 = map(float, range_vals)
        return self.rescale()

    @property
    def range(self) -> list[int | float]:
        return [self._r0, self._r1]

    def set_range_round(self, range_vals: list[int | range]) -> ScaleBand:
        """
        Sets the scale's range to the specified array of values
        and sets scale's interpolator to :code:`interpolate_round`.

        Parameters
        ----------
        range_vals : list[int | range]
            Range values

        Returns
        -------
        ScaleBand
            Itself
        """
        self._r0, self._r1 = map(float, range_vals)
        self._round = True
        return self.rescale()

    @property
    def bandwidth(self):
        return self._bandwidth

    @property
    def step(self):
        return self._step

    def set_round(self, round_val: bool) -> ScaleBand:
        """
        Enable or disable rounding accordingly

        Parameters
        ----------
        round_val : bool
            Round value

        Returns
        -------
        ScaleBand
            Itself
        """
        self._round = bool(round_val)
        return self.rescale()

    @property
    def round(self):
        return self._round

    def set_padding(self, padding: int | float) -> ScaleBand:
        """
        A convenience method for setting the inner and outer padding
        to the same padding value.

        Parameters
        ----------
        padding : int | float
            Padding value

        Returns
        -------
        ScaleBand
            Itself
        """
        self._padding_outer = float(padding)
        self._padding_inner = min(1, self._padding_outer)
        return self.rescale()

    @property
    def padding(self):
        return self._padding_inner

    def set_padding_inner(self, padding_inner: int | float) -> ScaleBand:
        """
        Sets the inner padding to the specified number which must
        be less than or equal to 1

        Parameters
        ----------
        padding_inner : int | float
            Inner padding value

        Returns
        -------
        ScaleBand
            Itself
        """
        self._padding_inner = min(1, float(padding_inner))
        return self.rescale()

    def set_padding_outer(self, padding_outer: int | float) -> ScaleBand:
        """
        Sets the outer padding to the specified number
        which is typically in the range [0, 1]

        Parameters
        ----------
        padding_outer : int | float
            Outer padding value

        Returns
        -------
        ScaleBand
            Itself
        """
        self._padding_outer = float(padding_outer)
        return self.rescale()

    @property
    def padding_inner(self):
        return self._padding_inner

    @property
    def padding_outer(self):
        return self._padding_outer

    def set_align(self, align: int | float) -> ScaleBand:
        """
        Sets the alignment to the specified value which must
        be in the range [0, 1]

        Parameters
        ----------
        align : int | float
            Alignment value

        Returns
        -------
        ScaleBand
            Itself
        """
        self._align = max(0, min(1, float(align)))
        return self.rescale()

    @property
    def align(self):
        return self._align

    def copy(self):
        return (
            ScaleBand()
            .set_domain(self._domain)
            .set_range([self._r0, self._r1])
            .set_round(self._round)
            .set_padding_inner(self._padding_inner)
            .set_padding_outer(self._padding_outer)
            .set_align(self._align)
        )


@overload
def scale_band() -> ScaleBand: ...


@overload
def scale_band(range_vals: list[int | float]) -> ScaleBand: ...


@overload
def scale_band(domain: list[T], range_vals: list[int | float]) -> ScaleBand: ...


def scale_band(*args):
    """
    Builds a new band scale with the specified domain
    and range, no padding, no rounding and center alignment

    Parameters
    ----------
    domain : list[T]
        Array of values
    range_vals : list[int | float]
        Array of numbers

    Returns
    -------
    ScaleBand
        Scale object

    Examples
    --------

    >>> d3.scale_band(["a", "b", "c"], [0, 960])
    """
    scale = ScaleBand()
    if len(args) == 1:
        return init_range(scale, range_vals=args[0])
    elif len(args) == 2:
        domain, range_vals = args
        return init_range(scale, domain=domain, range_vals=range_vals)
    return init_range(scale)


@overload
def scale_point() -> ScaleBand: ...


@overload
def scale_point(range_vals: list[int | float]) -> ScaleBand: ...


@overload
def scale_point(domain: list[T], range_vals: list[int | float]) -> ScaleBand: ...


def scale_point(*args):
    """
    Builds a new point scale with the specified domain and
    range, no padding, no rounding and center alignment

    Parameters
    ----------
    domain : list[T]
        Array of values
    range_vals : list[int | float]
        Array of numbers

    Returns
    -------
    ScaleBand
        Scale object

    Examples
    --------

    >>> d3.scale_point(["a", "b", "c"], [0, 960])
    """
    scale = ScaleBand().set_padding_inner(1)
    if len(args) == 1:
        return init_range(scale, range_vals=args[0])
    elif len(args) == 2:
        domain, range_vals = args
        return init_range(scale, domain=domain, range_vals=range_vals)
    return init_range(scale)



from inspect import signature


def text_constant(value):
    def callback(node, data, i, group):
        node.text = value

    return callback


def text_function(value):
    nargs = len(signature(value).parameters)

    def callback(node, data, i, group):
        args = [data, i, group][:nargs]
        node.text = value(*args)

    return callback



from collections.abc import Callable
from typing import TypeVar

T = TypeVar("T")


def interpolate_discrete(range_: list[T]) -> Callable[[float], T]:
    """
    Returns a discrete interpolator for the given array of values.

    Parameters
    ----------
    range_ : list[T]
        List of elements

    Returns
    -------
    Callable[[float], T]
        Interpolator
    """
    n = len(range_)
    return lambda t: range_[max(0, min(n - 1, int(t * n)))]
